# -*- coding: utf-8 -*-
import csv, os
import tkinter as tk
from tkinter.filedialog import askopenfilename
from itertools import count
import xlrd
from openpyxl import load_workbook, Workbook

class File:
    def __init__(self, filepath):
        self.filepath = filepath
        self.filename = os.path.basename(filepath)
        self.data = None
        
        
    @classmethod
    def new(cls, data, filename, directory=None, overwrite=True, includeHdr=True):
        """
        directory will default to downloads folder if left blank
        """
        if directory is None:
            directory = cls.download_dir()
            pass
        filepath = os.path.join(directory, filename)
        if os.path.exists(filepath) and not overwrite:
            raise Exception('This path exists already. Set overwrite true if you want to overwrite')
        obj=cls(filepath)
        obj.data=data
        obj.export_csv(filepath, data, includeHdr=includeHdr)
        return obj
        
    
    @staticmethod
    def download_dir():
        return os.path.join(os.getenv('USERPROFILE'), 'Downloads')

    
    @staticmethod
    def remap(fieldnames):
        uq={col:count(1) for col in list(set(fieldnames))}
        done=[]
        new=[]
        for col in fieldnames:
            ct=""
            if col in done:
                ct="-{num}".format(num=next(uq[col]))
            else:
                done.append(col)
            new.append('{column}{count}'.format(column=col, count=ct))
        return new
    
    
    

    ### CSV
    @staticmethod
    def load_csv(filepath, encoding='utf-8', delim=","):
        with open(filepath, encoding=encoding) as f:
            reader=csv.reader(f, delimiter=delim)
            fieldnames=File.remap(next(reader))
            data=[dict(zip(fieldnames,row)) for row in reader]
        return data
    
    @staticmethod
    def as_text(filepath, encoding='utf-8'):
        with open(filepath, 'r', encoding=encoding) as f:
            text = f.read()
        return text
    
    @staticmethod
    def load_tsv(filepath, encoding='utf-8', colmask=0):
        with open(filepath, encoding=encoding) as f:
            text = f.read()
            data = [[i for i in x.split('\t')] for x in text.split('\n')]
            hdrs=data[0]
            if colmask>0:
                hdrs=hdrs[:colmask]
            hdrs=File.remap(hdrs)
            func=lambda x: True if colmask==0 else len(x)>=colmask
            data = [dict(zip(hdrs,x[:colmask]))for x in data[1:] if func(x)]
        return data
    
    
    @staticmethod
    def export_csv(filepath, data, includeHdr=True):
        with open(filepath, 'w', newline='') as f:
            writer=csv.DictWriter(f, list(data[0].keys()))
            if includeHdr:
                writer.writeheader()
            writer.writerows(data)
         
            
    @staticmethod
    def read_xl(filepath):
        if str(filepath).lower().endswith('.xls'):
            wb=xlrd.open_workbook(filepath)
            sheets = wb.sheets()
            output = {}
            for sheet in sheets:
                output[sheet.name]=[sheet.row_values(x) for x in range(sheet.nrows)]
        else:
            wb=load_workbook(filepath, read_only=True,
                             data_only=True, keep_links=False)
            output = {}
            for sheet in wb.worksheets:
                output[sheet.title]=[x for x in sheet.values]
        return output
    
    @staticmethod
    def create_xls(data, title, filepath):
        wb=Workbook()
        ws = wb.active
        ws.title = title
        fields = list(data[0].keys())
        ws.append(fields)
        for row in data:
            ws.append(list(row.values()))
        wb.save(filepath)
        wb.close()
        file = File(filepath)
        file.data = data
        return file

            
    @staticmethod
    def find_file(title:str='', message:str=None):
        msg = f"File Finder: {title}"
        msg += f"\nPlease hold ALT and press TAB until you find the window titled '{title}'"
        if message is None:
            msg += f"\n{message}"
        print(msg)
        root=tk.Tk()
        root.withdraw()
        filepath = askopenfilename(title=title)
        root.destroy()
        return filepath


    ### GENERAL
    @staticmethod
    def delete(filepath):
        try:
            os.remove(filepath)
        except:
            print('File does not exist ({})'.format(filepath))    
            
            
    @staticmethod
    def listdir(folderpath):
        files=os.listdir(folderpath)
        return [File(os.path.join(folderpath,x)) for x in files]
    
    
    def __repr__(self):
        return self.filepath


from pathlib import Path
from anyio import Path as ioPath
import os, asyncio
import base64

class FileAlreadyExists(Exception):
    ...

class AsyncTempFile():
    def __init__(self,
                 filepath:str|ioPath|os.PathLike,
                 mode:str="wb"):
        self.filepath = (filepath 
                         if isinstance(filepath, ioPath) 
                         else ioPath(str(filepath)))
        self.mode = mode
        self.mgr  = None
        self.file = None
        
    async def __aenter__(self):
        self.mgr = await self.filepath.open(
            mode=self.mode
        )
        self.file = await self.mgr.__aenter__()
        return self.file
        
    async def __aexit__(self, ex_tp, ex_v, tx_tb):
        if self.mgr is not None:
            await self.mgr.__aexit__(ex_tp, ex_v, tx_tb)
        
        
async def process_async_temp_file(filepath:str|ioPath|os.PathLike, raw:bytes, processor:object):
    """
    Arguments:
    - filepath (str or pathlike object; absolute/relative path to the desired temporary location; it will check to make sure the path does not exist before running)
    - raw (bytes: this is the raw content of the file in bytes)
    - processor (sync or async callable function that takes one argument 'filepath')
    
    Utilizes AsyncTempFile to 'process' the file using this strategy:
    1. Write/Create temporary file by writing 'raw' (bytes) to the provided 'filepath'
    2. Run the processor method using the temporary filepath as the only argument
        - the result of 'processor' is returned at the end of the function
    3. Delete the file from file system
    """
    
    # raw is base64 encoded
    data = base64.b64decode(raw)
    filepath = ioPath(str(filepath))
    if await filepath.exists():
        raise FileAlreadyExists(f"Temporary file already exists: '{str(filepath)}'")
    if not callable(processor):
        raise Exception("Processor must be a callable function or a class with a __call__ method defined.")
    async with AsyncTempFile(filepath) as file:
        await file.write(data)
    if asyncio.iscoroutinefunction(processor):
        rsp = await processor(filepath)
    else:
        rsp = processor(filepath)
    os.remove(filepath)
    return rsp


def process_temp_file(filepath:str|ioPath|os.PathLike, raw:bytes, processor:object):
    """
    Arguments:
    - filepath (str or pathlike object; absolute/relative path to the desired temporary location; it will check to make sure the path does not exist before running)
    - raw (bytes: this is the raw content of the file in bytes)
    - processor (callable function that takes one argument 'filepath')
    
    Utilizes os.open to 'process' the file using this strategy:
    1. Write/Create temporary file by writing 'raw' (bytes) to the provided 'filepath'
    2. Run the processor method using the temporary filepath as the only argument
        - the result of 'processor' is returned at the end of the function
    3. Delete the file from file system
    """
    filepath = Path(str(filepath))
    if filepath.exists():
        raise FileAlreadyExists(f"Temporary file already exists: '{str(filepath)}'")
    if not callable(processor):
        raise Exception("Processor must be a callable function or a class with a __call__ method defined.")
    with open(filepath, 'wb') as file:
        file.write(raw)
    rsp = processor(filepath)
    os.remove(filepath)
    return rsp



from .collection import Collection, fast_group
from .converters import (
    string_to_datetime, string_to_int, string_to_money,
    truncate, TableMapper
)
from .file import File
from .key_gen import (
    gen_alpha_key, gen_alphanumeric_key,
    gen_numeric_key, gen_power_key
)
from .point import DataPoint
from .temp_file import (
    AsyncTempFile, process_async_temp_file, 
    process_temp_file
    )


# -*- coding: utf-8 -*-
import string, random

def gen_alphanumeric_key(length:int=24, case_sensitive:bool=True):
    """Generate a randomized key using numbers and letters"""
    allowed = list(string.ascii_letters + string.digits)
    if not case_sensitive:
        allowed = list(set([str(x).upper() for x in allowed]))
    return ''.join(random.choice(allowed) for i in range(length))

def gen_alpha_key(length:int=24, case_sensitive:bool=True):
    """Generate a randomized key using only letters"""
    allowed = list(string.ascii_letters)
    if not case_sensitive:
        allowed = list(set([str(x).upper() for x in allowed]))
    return ''.join(random.choice(allowed) for i in range(length))

def gen_numeric_key(length:int=24):
    """Generate a randomized key using only numbers"""
    allowed = list(string.digits)
    return ''.join(random.choice(allowed) for i in range(length))


def gen_power_key(length, case_sensitive=True, include_chars="", exclude_chars=""):
    allowed = list(string.ascii_letters + string.digits)
    if not case_sensitive:
        allowed = [char.upper() for char in allowed]
    allowed = [char for char in allowed if char not in exclude_chars]
    allowed += [char for char in include_chars if char not in allowed]
    if isinstance(length, tuple):
        min_length, max_length = length
        length = random.randint(min_length, max_length)
    elif isinstance(length, int):
        length = length
    return ''.join(random.choice(allowed) for _ in range(length))





class Collection(list):
    def __init__(self, data):
        super(Collection, self).__init__(data)
        
        
    @classmethod
    def fromListOfLists(cls, data:list, hdrIx:int=0, func=None):
        hdr = data[hdrIx]
        data = [dict(zip(hdr,x)) for x in data[hdrIx:]]
        if func is None:
            func=lambda x: True
        data = [x for x in data if func(x)]
        return cls(data)
    
    
    @staticmethod
    def listdict_to_listlist(data:list):
        out = []
        hdr = list(data[0].keys())
        out.append(hdr)
        for r in data:
            s=list(r.values())
            out.append(s)
        return out
    
    @staticmethod
    def chunkify(items:list, chunk=1_000):
        """"Takes an input list and returns a list of lists 
        chunked according to chunk size.
        """
        total = len(items)
        remaining = total
        A = 0
        Z = min(A + chunk, total)
        results = []
        while remaining > 0:
            Z = min(total, A + chunk)
            results.append(items[A:Z])
            remaining = total - Z
            A = Z 
        return results
    
    @classmethod
    def empty(cls):
        return cls([])
    
    
    @classmethod
    def where(cls, iterable, func):
        return cls.query(iterable, func)
    
    
    @staticmethod
    def query(iterable, func):
        t=[x for x in iterable if func(x)]
        return Collection(t)
    
    
    def align(self):
        # align a list of dicts so that all keys have same order 
        # this will base everything off of the first dict item in the list
        template = list(self[0].keys())
        for i in range(len(self)):
            x = self[i]
            r = {k:x.get(k,None) for k in template}
            self[i]=r
    
    
    @staticmethod
    def get_header(data):
        top=list(data[0].keys())
        return top
    
    
    @property
    def headers(self):
        return self.get_header(self)
    
    
    
    def sum_if(self, sumField:str, filterFunc=None, decimals:int=0):
        return self._sum_if(self, sumField, filterFunc, decimals)
    
    
    @staticmethod
    def _sum_if(iterable, sumField, filterFunc=None, decimals:int=0):
        if filterFunc is None:
            filterFunc = lambda x: True
        rsp = round(sum([(x.get(sumField,0) or 0) for x in iterable if filterFunc(x)]), decimals)
        if decimals==0:
            rsp=int(rsp)
        return rsp
    
    
    def max_if(self, maxField, filterFunc=None):
        return self._max_if(self, maxField, filterFunc)
    
    
    @staticmethod
    def _max_if(iterable, maxField, filterFunc=None):
        if filterFunc is None:
            filterFunc = lambda x: True
        m = [x[maxField] for x in iterable if filterFunc(x)]
        if len(m)==0:
            return None
        return max(m)
    
    
    def min_if(self, minField, filterFunc=None):
        return self._min_if(self, minField, filterFunc)
    
    
    @staticmethod
    def _min_if(iterable, minField, filterFunc=None):
        if filterFunc is None:
            filterFunc = lambda x: True
        m = [x[minField] for x in iterable if filterFunc(x)]
        if len(m)==0:
            return None
        return min(m)
    
    
    
    def count_if(self, filterFunc=None):
        if filterFunc is None:
            filterFunc = lambda x: True
        rsp = len([x for x in self if filterFunc(x)])
        return rsp
    
    
    def search(self, matchValue, matchField:str):
        rsp = [x for x in self if x[matchField]==matchValue]
        return Collection(rsp)
    
    
    def index_match(self, other, matchValue, matchField):
        other.sort()
        try:
            rsp = other.search(matchValue, matchField)
        except:
            other = Collection(other)
            rsp = other.search(matchValue, matchField)
        return rsp
    
    
    def order_by(self, by:str, reverse=False):
        self.sort(key=lambda x: x[by], reverse=reverse)
        



def fast_group(iterable, field, save_iterable=False, dotget=False):
    """
    iterable = list of dictionaries (or a collection object)
    field = str or list of str (for isolated multi-index grouping (1st and 2nd indexes can repeat, but their combinations cannot)
    save_iterable = bool (if true, retain the originally passed iterable to avoid accidental deletion, otherwise, it trashes itself)
    dotget = bool (if true, use a "." (dot) notation when filtering vs ["key"] (hash) notation (like a normal dictionary))
    """
    if save_iterable:
        runner = iterable.copy()
    else:
        runner = iterable
    out = {}
    multi = not isinstance(field, str)
    get = lambda item, field: item[field] if not dotget else getattr(item, field)
    while len(runner)>0:
        r = runner.pop()
        if multi:
            key = tuple(get(r, i) for i in field)
        else:
            key = get(r, field)
        c = out.get(key, Collection.empty())
        c.append(r)
        out[key]=c
    return out


from copy import deepcopy

class DataPoint:
    def __init__(self, row, datapoint_name=None):
        self.row = row
        self.og  = deepcopy(row)
        self.extra = {}
        self.datapoint_name = datapoint_name
    
    def __getitem__(self, k):
        if k not in self.row:
            return self.extra.get(k, None)
        return self.row[k]
    
    def __setitem__(self, k, v):
        if k not in self.row:
            self.extra[k] = v
        else:
            self.row[k] = v
    
    def dict(self):
        return deepcopy(self.row)

    def save(self):
        self.og = deepcopy(self.row)
    
    def rollback(self):
        self.row = deepcopy(self.og)
    
    @property
    def has_changes(self):
        return any([self.og[k]!=v for k,v in self.row.items()])
    
    def __repr__(self):
        if self.datapoint_name is not None:
            return f"<DataPoint ({self.datapoint_name})>"
        return "<DataPoint>"


import dateutil.parser
import json

def string_to_datetime(datestring):
    yourdate = dateutil.parser.parse(datestring)
    return yourdate

def string_to_money(moneystring:str) -> float:
    if isinstance(moneystring, (int,float)):
        return round(float(moneystring), 2)
    og = moneystring
    rms = [("$",""), ("(", "-"), (")",""), (",",""), (" ","")]
    for wrong, right in rms:
        moneystring = moneystring.replace(wrong, right)
    try:
        return round(float(moneystring or "0"), 2)
    except:
        raise Exception(f"{og} is not a valid money data type.")
    
def string_to_int(intstr:str) -> int:
    if isinstance(intstr, (int,float)):
        return int(round(intstr, 0))
    try:
        return int(round(float(intstr or '0')))    
    except:
        raise Exception(f"{intstr} is not a valid integer")
    
def truncate(f, n) -> float:
    """Round a float to the n-th decimal point without rounding up or down (truncate it)"""
    s = '{}'.format(f)
    if 'e' in s or 'E' in s:
        return '{0:.{1}f}'.format(f, n)
    i, p, d = s.partition('.')
    return float('.'.join([i, (d+'0'*n)[:n]]))


class TableMapper():
    def __init__(self, 
                 mapp:dict[str, list], 
                 case_sensitive:bool=True,
                 auto_meta:bool=False,
                 skip_missing:bool=False,
                 rem_chars:list=None
                 ):
        self.case_sensitive = case_sensitive
        self.auto_meta = auto_meta
        self.skip_missing = skip_missing
        self.rem_chars = rem_chars
        self.mapp = {
            self.standardize_key(key):
                [self.standardize_key(val) for val in vals]
            for key, vals in mapp.items()
        }
    
    def nocase(self, value):
        return str(value).lower()
    
    def standardize_key(self, value:str):
        if not self.case_sensitive:
            value = str(value).lower()
        if isinstance(self.rem_chars,list):
            for ch in self.rem_chars:
                value = str(value).replace(ch,'')
        return value
    
    def remap(self, remapper:dict, row:dict):
        new = {
            newk:row[oldk] for oldk,newk in remapper.items()
        }
        if self.auto_meta:
            new['_meta_'] = json.dumps(row, default=str, indent=2)
        return new
    
    def process(self, data:list[dict]):
        r = data[0]
        remapper = {}
        mapped = []
        remaining = list(self.mapp.keys())
        for key in r.keys():
            test = self.standardize_key(key)
            if test in self.mapp and test not in mapped:
                remapper[key]=test # key stays the same
                mapped.append(test)
                remaining.remove(test)
                continue
            for newkey,vs in self.mapp.items():
                if newkey in mapped:
                    continue # no duplicates
                if test in vs:
                    remapper[key]=newkey
                    mapped.append(newkey)
                    remaining.remove(newkey)
                    break
        if len(remaining) > 0 and not self.skip_missing:
            raise Exception("Some Fields are missing from the table: {}".format(", ".join(remaining)))
        output = list(map(lambda row: self.remap(remapper, row), data))
        return output